/* $Id: isaac_img_zpoint.c,v 1.56 2010/03/02 13:26:12 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2010/03/02 13:26:12 $
 * $Revision: 1.56 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_calib.h"
#include "irplib_strehl.h"
#include "irplib_stdstar.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define PHOT_STAR_RADIUS                30.0
#define PHOT_BACKGROUND_R1              40.0
#define PHOT_BACKGROUND_R2              60.0

#define DEF_LOCATE_SX                   10
#define DEF_LOCATE_SY                   10

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_img_zpoint_create(cpl_plugin *);
static int isaac_img_zpoint_exec(cpl_plugin *);
static int isaac_img_zpoint_destroy(cpl_plugin *);
static int isaac_img_zpoint(cpl_parameterlist *, cpl_frameset *);
static cpl_table * isaac_img_zpoint_reduce(cpl_frameset *, const char *,
        const char *, const char *, const char *, const char *, cpl_image **);
static cpl_imagelist * isaac_img_zpoint_load(cpl_frameset *, const char *, 
        const char *, const char *, const char *);
static int isaac_img_zpoint_save(cpl_table *, cpl_image *, cpl_frameset *, 
        cpl_parameterlist *, cpl_frameset *);
static cpl_table * isaac_img_zpoint_photom(cpl_imagelist *, cpl_bivector *);
static int isaac_img_zpoint_get_mag(const char *, double, double, isaac_band);
static int isaac_img_zpoint_gradients(cpl_vector *, cpl_bivector *, double *, 
        double *, double *, double *);
static cpl_bivector * isaac_img_zpoint_get_offsets(cpl_frameset *);
static int isaac_img_zpoint_compute_keywords(cpl_frameset *, double *,double *);
static cpl_image * isaac_img_zpoint_check_im(cpl_imagelist *,
        cpl_bivector *, double, double, double);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double      ra;
    double      dec;
    double      magnitude;
    int         sx;
    int         sy;
    double      phot_star_radius;
    double      phot_bg_r1;
    double      phot_bg_r2;
    int         chopping;
    int         check_im;
    /* Outputs */
    double      dit;
    char        filter[512];
    isaac_band  band;
    char        starname[512];
    char        sptype[512];
    char        catalog[512];
    double      zpoint;
    double      zpointrms;
    double      flux_med;
    double      gradx;
    double      grady;
    double      graddx;
    double      graddy;
    double      fwhm_mean;
} isaac_img_zpoint_config;

static char isaac_img_zpoint_description[] = 
"isaac_img_zpoint -- Zero point recipe\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "ISAAC_IMG_ZPOINT_RAW" or\n"
"raw-file.fits "ISAAC_IMG_ZPOINT_CHOPPING_RAW" or\n"
"stdstars.fits "ISAAC_CALIB_STDSTARS" or\n"
"flat-file.fits "ISAAC_CALIB_FLAT" or\n"
"detlin-a-file.fits "ISAAC_CALIB_DETLIN_A" or\n"
"detlin-b-file.fits "ISAAC_CALIB_DETLIN_B" or\n"
"detlin-c-file.fits "ISAAC_CALIB_DETLIN_C"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_img_zpoint",
                    "Zero point computation recipe",
                    isaac_img_zpoint_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_img_zpoint_create,
                    isaac_img_zpoint_exec,
                    isaac_img_zpoint_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return -1;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --star_r */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.star_r",CPL_TYPE_DOUBLE,
            "the star radius", "isaac.isaac_img_zpoint", PHOT_STAR_RADIUS);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "star_r");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --bg_r1 */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.bg_r1", CPL_TYPE_DOUBLE,
            "the internal background radius", "isaac.isaac_img_zpoint", 
            PHOT_BACKGROUND_R1);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "bg_r1");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --bg_r2 */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.bg_r2", CPL_TYPE_DOUBLE,
            "the external background radius", "isaac.isaac_img_zpoint", 
            PHOT_BACKGROUND_R2);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "bg_r2");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --ra */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.ra", CPL_TYPE_DOUBLE,
            "RA in degrees", "isaac.isaac_img_zpoint", 999.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "ra");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --dec */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.dec", CPL_TYPE_DOUBLE,
            "DEC in degrees", "isaac.isaac_img_zpoint", 999.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "dec");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --mag */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.mag", CPL_TYPE_DOUBLE,
            "magnitude", "isaac.isaac_img_zpoint", 99.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "mag");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --sx */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.sx", CPL_TYPE_INT, 
            "x size of the search window", "isaac.isaac_img_zpoint",
            DEF_LOCATE_SX);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "sx");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --sy */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.sy", CPL_TYPE_INT, 
            "y size of the search window", "isaac.isaac_img_zpoint",
            DEF_LOCATE_SY);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "sy");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --check_im */
    p = cpl_parameter_new_value("isaac.isaac_img_zpoint.check_im",
            CPL_TYPE_BOOL, "flag to create the check image",
            "isaac.isaac_img_zpoint", FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "check_im");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return -1;

    return isaac_img_zpoint(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return -1;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_parameter   *   par;
    const char      *   stdstars;
    const char      *   flat;
    const char      *   detlin_a;
    const char      *   detlin_b;
    const char      *   detlin_c;
    cpl_frameset    *   rawframes;
    cpl_table       *   tab;
    cpl_image       *   check_im;
    
    /* Initialise */
    rawframes = NULL;
    par = NULL;
    isaac_img_zpoint_config.starname[0] = (char)0;
    isaac_img_zpoint_config.sptype[0] = (char)0;
    isaac_img_zpoint_config.filter[0] = (char)0;
    isaac_img_zpoint_config.catalog[0] = (char)0;
    isaac_img_zpoint_config.gradx = -1.0;
    isaac_img_zpoint_config.grady = -1.0;
    isaac_img_zpoint_config.graddx = -1.0;
    isaac_img_zpoint_config.graddy = -1.0;
    isaac_img_zpoint_config.fwhm_mean = -1.0;

    /* Retrieve input parameters */
    /* --ra */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.ra");
    isaac_img_zpoint_config.ra = cpl_parameter_get_double(par);
    /* --dec */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.dec");
    isaac_img_zpoint_config.dec = cpl_parameter_get_double(par);
    /* --mag */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.mag");
    isaac_img_zpoint_config.magnitude = cpl_parameter_get_double(par);
    /* --sx */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.sx");
    isaac_img_zpoint_config.sx = cpl_parameter_get_int(par);
    /* --sy */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.sy");
    isaac_img_zpoint_config.sy = cpl_parameter_get_int(par);
    /* --star_r */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.star_r");
    isaac_img_zpoint_config.phot_star_radius = cpl_parameter_get_double(par);
    /* --bg_r1 */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.bg_r1");
    isaac_img_zpoint_config.phot_bg_r1 = cpl_parameter_get_double(par);
    /* --bg_r2 */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.bg_r2");
    isaac_img_zpoint_config.phot_bg_r2 = cpl_parameter_get_double(par);
    /* --check_im */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_zpoint.check_im");
    isaac_img_zpoint_config.check_im = cpl_parameter_get_bool(par);

    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return -1;
    }

    /* Retrieve calibration data */
    detlin_a = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_A);
    detlin_b = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_B);
    detlin_c = isaac_extract_filename(framelist, ISAAC_CALIB_DETLIN_C);
    flat     = isaac_extract_filename(framelist, ISAAC_CALIB_FLAT);
    stdstars = isaac_extract_filename(framelist, ISAAC_CALIB_STDSTARS);

    /* Retrieve raw frames */
    if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_ZPOINT_RAW)) != NULL) {
        isaac_img_zpoint_config.chopping = 0;
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_ZPOINT_CHOPPING_RAW)) != NULL) {
        isaac_img_zpoint_config.chopping = 1;
    } else {
        cpl_msg_error(cpl_func, "Cannot find raw frames in the input list");
        return -1;
    }

    /* Compute the zpoint values */
    cpl_msg_info(cpl_func, "Reduce the data");
    cpl_msg_indent_more();
    if ((tab = isaac_img_zpoint_reduce(rawframes, stdstars, flat, detlin_a, 
                    detlin_b, detlin_c, &check_im)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot reduce the data");
        cpl_frameset_delete(rawframes);
        cpl_msg_indent_less();
        return -1;
    }
    cpl_msg_indent_less();

    /* Save the products */
    cpl_msg_info(cpl_func, "Save the paf file");
    cpl_msg_indent_more();
    if (isaac_img_zpoint_save(tab, check_im, rawframes, parlist, 
                framelist) == -1) {
        cpl_msg_error(cpl_func, "Cannot save the paf file");
        if (check_im) cpl_image_delete(check_im);
        cpl_frameset_delete(rawframes);
        cpl_table_delete(tab);
        cpl_msg_indent_less();
        return -1;
    }
    cpl_frameset_delete(rawframes);
    cpl_msg_indent_less();

    /* Free and return */
    if (check_im) cpl_image_delete(check_im);
    cpl_table_delete(tab);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    set         the set of frames
  @param    stdstars    the standard stars catalog or NULL
  @param    flat    the flat field name or NULL 
  @param    detlin_a    the a coeffs for linearity or NULL
  @param    detlin_b    the b coeffs for linearity or NULL
  @param    detlin_c    the c coeffs for linearity or NULL
  @param    check_im    the produced check image or NULL
  @return   a cpl_table with the results or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_img_zpoint_reduce(
        cpl_frameset    *   set,
        const char      *   stdstars,
        const char      *   flat,
        const char      *   detlin_a,
        const char      *   detlin_b,
        const char      *   detlin_c,
        cpl_image       **  check_im)
{
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    const char          *   sval;
    cpl_imagelist       *   imlist;
    cpl_image           *   ima;
    cpl_mask            *   kernel;
    int                     niter;
    int                     size_x, size_y;
    double                  pos_x, pos_y, pos_x_cen, pos_y_cen, dist, min_dist;
    cpl_apertures       *   aperts;
    cpl_bivector        *   offsets,
                        *   positions;
    double                  off_x, off_y;
    int                     llx, lly, urx, ury;
    double                  val, sqsum, zprms, avg_zp;
    cpl_table           *   tab_res;
    cpl_vector          *   tmp_vec;
    int                     nb_ok;
    int                     i;

    /* Initialise */
    *check_im = NULL;
    pos_x_cen = pos_y_cen = -1.0;

    /* Get the filter name, DIT, RA and DEC  */
    cur_frame = cpl_frameset_get_frame(set, 0);
    plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    if ((sval = isaac_pfits_get_filter(plist)) == NULL) {
        cpl_propertylist_delete(plist);
        cpl_ensure(0, cpl_error_get_code(), NULL);
    }
    (void)strncpy(isaac_img_zpoint_config.filter, sval, 512);
    isaac_img_zpoint_config.filter[511] = (char)0;
    if (isaac_img_zpoint_config.ra > 998.0) 
        isaac_img_zpoint_config.ra = isaac_pfits_get_ra(plist);
    if (isaac_img_zpoint_config.dec > 998.0) 
        isaac_img_zpoint_config.dec = isaac_pfits_get_dec(plist);
    isaac_img_zpoint_config.dit = isaac_pfits_get_dit(plist);
    cpl_propertylist_delete(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get some header informations");
        return NULL;
    }
    cpl_msg_info(cpl_func, "Using star at position: RA = %g; DEC = %g", 
            isaac_img_zpoint_config.ra, isaac_img_zpoint_config.dec);
   
    /* Get the band */
    if ((isaac_img_zpoint_config.band = 
                isaac_get_bbfilter(isaac_img_zpoint_config.filter)) ==
            ISAAC_BAND_UNKNOWN) {
        cpl_msg_error(cpl_func, "Cannot associate the filter to a BB one");
        return NULL;
    }
    
    /* Get the standard star information from database */
    if (isaac_img_zpoint_config.magnitude > 98.0) {
        cpl_msg_info(cpl_func, "Get the star magnitude");
        if (isaac_img_zpoint_get_mag(stdstars, isaac_img_zpoint_config.ra,
                                    isaac_img_zpoint_config.dec,
                                    isaac_img_zpoint_config.band) == -1) {
            cpl_msg_error(cpl_func,"Cannot get the magnitude from the catalog");
            return NULL;
        }
    }
    cpl_msg_info(cpl_func, "Star magnitude with filter %s : %g",
            isaac_img_zpoint_config.filter,
            isaac_img_zpoint_config.magnitude);
    
    /* Load the images */
    cpl_msg_info(cpl_func, "Load images");
    if ((imlist = isaac_img_zpoint_load(set, flat, detlin_a, detlin_b,
                    detlin_c)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the images"); 
        return NULL;
    }
       
    /* Detect the central object in the first frame */
    cpl_msg_info(cpl_func, "Detect a bright object in the first frame");
    size_x = cpl_image_get_size_x(cpl_imagelist_get(imlist, 0));
    size_y = cpl_image_get_size_y(cpl_imagelist_get(imlist, 0));
    kernel = cpl_mask_new(3, 3);
    cpl_mask_not(kernel);
    ima = cpl_image_new(size_x, size_y,
                        cpl_image_get_type(cpl_imagelist_get(imlist, 0)));
    cpl_image_filter_mask(ima, cpl_imagelist_get(imlist, 0), kernel,
                          CPL_FILTER_MEDIAN, CPL_BORDER_FILTER);
    cpl_mask_delete(kernel);
    aperts = cpl_apertures_extract_sigma(ima, 5.0);
    cpl_image_delete(ima);
    min_dist = size_x * size_x + size_y * size_y;
    for (i=0; i<cpl_apertures_get_size(aperts); i++) {
        pos_x = cpl_apertures_get_centroid_x(aperts, i+1);
        pos_y = cpl_apertures_get_centroid_y(aperts, i+1);
        dist = (pos_x-size_x/2)*(pos_x-size_x/2) +
               (pos_y-size_y/2)*(pos_y-size_y/2);
        if (dist<min_dist) {
            min_dist = dist;
            pos_x_cen = pos_x;
            pos_y_cen = pos_y;
        }
    }
    cpl_apertures_delete(aperts);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot find the central object");
        cpl_imagelist_delete(imlist);
        return NULL;
    }
    cpl_msg_info(cpl_func, "Bright object position: %g %g", 
            pos_x_cen, pos_y_cen);
   
    /* Get the offsets */
    cpl_msg_info(cpl_func, "Read the offsets in the header");
    offsets = isaac_img_zpoint_get_offsets(set);
    
    /* Deduce the objects positions  */
    niter = cpl_imagelist_get_size(imlist);
    positions   = cpl_bivector_new(niter);
    for (i=0; i<niter; i++) {
        off_x = cpl_vector_get(cpl_bivector_get_x(offsets), i) -
            cpl_vector_get(cpl_bivector_get_x(offsets), 0);
        off_y = cpl_vector_get(cpl_bivector_get_y(offsets), i) -
            cpl_vector_get(cpl_bivector_get_y(offsets), 0);
        pos_x = pos_x_cen + off_x;
        pos_y = pos_y_cen + off_y;
        cpl_vector_set(cpl_bivector_get_x(positions), i, pos_x);
        cpl_vector_set(cpl_bivector_get_y(positions), i, pos_y);
    }
    cpl_bivector_delete(offsets);

    /* Refine the positions */
    cpl_msg_info(cpl_func, "Refine the star positions");
    for (i=1; i<niter; i++) {
        pos_x = cpl_vector_get(cpl_bivector_get_x(positions), i);
        pos_y = cpl_vector_get(cpl_bivector_get_y(positions), i);
        llx = pos_x - isaac_img_zpoint_config.sx;
        urx = pos_x + isaac_img_zpoint_config.sx;
        lly = pos_y - isaac_img_zpoint_config.sy;
        ury = pos_y + isaac_img_zpoint_config.sy;
        ima = cpl_imagelist_get(imlist, i);
        pos_x = cpl_image_get_centroid_x_window(ima, llx, lly, urx, ury);
        pos_y = cpl_image_get_centroid_y_window(ima, llx, lly, urx, ury);
        cpl_vector_set(cpl_bivector_get_x(positions), i, pos_x);
        cpl_vector_set(cpl_bivector_get_y(positions), i, pos_y);
    }
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot refine the positions");
        cpl_imagelist_delete(imlist);
        cpl_bivector_delete(positions);
        return NULL;
    }

    /* Create the check image if requested */
    if (isaac_img_zpoint_config.check_im) {
        *check_im = isaac_img_zpoint_check_im(imlist, positions, 
                isaac_img_zpoint_config.phot_star_radius, 
                isaac_img_zpoint_config.phot_bg_r1,
                isaac_img_zpoint_config.phot_bg_r2);
    }

    /* Reduce the images */
    cpl_msg_info(cpl_func, "Compute the photometry");
    cpl_msg_indent_more();
    if ((tab_res = isaac_img_zpoint_photom(imlist, positions)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot reduce");
        cpl_bivector_delete(positions);
        cpl_imagelist_delete(imlist);
        cpl_image_delete(*check_im);
        *check_im = NULL;
        cpl_msg_indent_less();
        return NULL;
    } 
    cpl_msg_indent_less();
    cpl_imagelist_delete(imlist);

    /* Compute the median of the flux */
    isaac_img_zpoint_config.flux_med =
        cpl_table_get_column_median(tab_res, "FLUX"); 

    /* Get the mean FWHM  */
    isaac_img_zpoint_config.fwhm_mean =
        cpl_table_get_column_mean(tab_res, "FWHMX");
    isaac_img_zpoint_config.fwhm_mean +=
        cpl_table_get_column_mean(tab_res, "FWHMY");
    isaac_img_zpoint_config.fwhm_mean /= 2.0;

    /* Compute the flux gradient */
    tmp_vec = cpl_vector_new(niter);
    for (i=0; i<niter; i++) {
        cpl_vector_set(tmp_vec, i, 
                cpl_table_get_double(tab_res, "FLUX", i, NULL));
    }
    isaac_img_zpoint_gradients(tmp_vec, positions,
            &(isaac_img_zpoint_config.gradx),
            &(isaac_img_zpoint_config.grady),
            &(isaac_img_zpoint_config.graddx),
            &(isaac_img_zpoint_config.graddy));
    cpl_vector_delete(tmp_vec);
    cpl_bivector_delete(positions);
    
    /* Compute the averages of the results for the zero point */
    tmp_vec = cpl_vector_new(niter);
    for (i=0; i<niter; i++) {
        cpl_vector_set(tmp_vec, i, 
                cpl_table_get_double(tab_res, "ZPOINT", i, NULL));
    }
    nb_ok = 0;
    avg_zp = sqsum = 0.0;
    if (isaac_img_zpoint_config.chopping == 0) {
        cpl_vector_sort(tmp_vec, 1);
        /* Reject highest and lowest value */
        for (i=1; i<niter-1; i++) {
            val = cpl_vector_get(tmp_vec, i);
            if (val > 0.0) {
                avg_zp += val;
                sqsum += val * val;
                nb_ok ++;
            }
        }
    } else if (isaac_img_zpoint_config.chopping == 1) {
        for (i=0; i<niter; i++) {
            val = cpl_vector_get(tmp_vec, i);
            if (val > 0.0) {
                avg_zp += val;
                sqsum += val * val;
                nb_ok ++;
            }
        }
    } else {
        cpl_msg_error(cpl_func, "unsupported mode"); 
        cpl_table_delete(tab_res);
        cpl_vector_delete(tmp_vec);
        cpl_image_delete(*check_im);
        *check_im = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_vector_delete(tmp_vec);
    if (nb_ok < 1) {
        cpl_msg_error(cpl_func, "no valid zpoint measurement: cannot compute"); 
        isaac_img_zpoint_config.zpoint    = -1.0;
        isaac_img_zpoint_config.zpointrms = -1.0;
    } else {
        avg_zp /= (double)nb_ok;
        sqsum  /= (double)nb_ok;
        zprms = sqsum - avg_zp * avg_zp;
        zprms = zprms > 0 ? sqrt(zprms) : 0;
        isaac_img_zpoint_config.zpoint    = avg_zp;
        isaac_img_zpoint_config.zpointrms = zprms;
    }

    /* Print final results */
    cpl_msg_info(cpl_func, "***** FINAL RESULTS *****");
    cpl_msg_info(cpl_func, "Zero point :   %g", isaac_img_zpoint_config.zpoint);
    cpl_msg_info(cpl_func, "Zero p. RMS:   %g", isaac_img_zpoint_config.zpointrms);
    
    return tab_res;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create an image with subimages and circles around the star
  @param    imlist      List of input images
  @param    positions   List of star positions
  @param    r1          First radius
  @param    r2          Second radius
  @param    r3          Third radius
  @return   The generated image

  The function takes sub images of the input images centered around the
  specified positions, and make a big images with the sub images next to
  each others. Circles are drawn around the first central position with
  three different radii (r1, r2, r3). 
 */
/*----------------------------------------------------------------------------*/
static cpl_image * isaac_img_zpoint_check_im(
        cpl_imagelist   *   imlist,
        cpl_bivector    *   positions,
        double              r1,
        double              r2,
        double              r3)
{
    int                     nima, in_nx, in_ny, nx, ny, box_sz, llx, lly;
    cpl_image           *   in_ima;
    float               *   pin_ima;
    cpl_image           *   out_ima;
    float               *   pout_ima;
    int                     out_pos, in_pos;
    double              *   pos_x;
    double              *   pos_y;
    double                  dist;
    int                     i, j, k;
 
    /* Check entries */
    if (imlist == NULL) return NULL;
    if (positions == NULL) return NULL;
    nima = cpl_imagelist_get_size(imlist);
    if (cpl_bivector_get_size(positions) != nima) return NULL;
    
    /* Initialise */
    in_ima = cpl_imagelist_get(imlist, 0);
    in_nx = cpl_image_get_size_x(in_ima);
    in_ny = cpl_image_get_size_y(in_ima);
    pos_x = cpl_bivector_get_x_data(positions);
    pos_y = cpl_bivector_get_y_data(positions);

    /* Create the output image */
    box_sz = 2 * (int)r3 + 1;
    nx = nima * box_sz;    
    ny = box_sz;
    out_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
    pout_ima = cpl_image_get_data_float(out_ima);

    /* Loop on the input images */
    for (k=0; k<nima; k++) {
        /* Get the current input image */
        in_ima = cpl_imagelist_get(imlist, k);
        pin_ima = cpl_image_get_data_float(in_ima);

        /* Get the sub image position */
        llx = (int)(pos_x[k] - r3);
        lly = (int)(pos_y[k] - r3);
        
        for (i=0; i<box_sz; i++) {
            for (j=0; j<box_sz; j++) {
                out_pos = (box_sz * k) + i + j * nx; 
                in_pos = llx + i + (lly+j) * in_nx;
                if (in_pos >= 0 && in_pos < in_nx*in_ny)
                    pout_ima[out_pos] = pin_ima[in_pos];
            }
        }
    }

    /* Draw circles */
    for (i=0; i<box_sz; i++) {
        for (j=0; j<box_sz; j++) {
            dist = sqrt((i - (box_sz/2)) * (i - (box_sz/2)) +
                (j - (box_sz/2)) * (j - (box_sz/2)));
            if ((fabs(dist-sqrt(r1*r1))) < 0.5) pout_ima[i + j * nx] = 10000;
            if ((fabs(dist-sqrt(r2*r2))) < 0.5) pout_ima[i + j * nx] = 10000;
            if ((fabs(dist-sqrt(r3*r3))) < 0.5) pout_ima[i + j * nx] = 10000;
        }
    }

    /* Return  */
    return out_ima;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the input images
  @param
  @return
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_img_zpoint_load(
        cpl_frameset    *   set,
        const char      *   flat,
        const char      *   detlin_a,
        const char      *   detlin_b,
        const char      *   detlin_c)
{
    cpl_imagelist       *   imlist;
    cpl_imagelist       *   imlist2;
    cpl_imagelist       *   diffs;
    cpl_image           *   ima;
    int                     niter;
    int                     i;
    
    /* Load the images */
    if (isaac_img_zpoint_config.chopping == 0) {
        if ((imlist = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 1,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            return NULL;
        }
    } else if (isaac_img_zpoint_config.chopping == 1) {
        if ((imlist = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 1,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            return NULL;
        }
        if ((imlist2 = cpl_imagelist_load_frameset(set, CPL_TYPE_FLOAT, 2,
                        0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the images"); 
            cpl_imagelist_delete(imlist);
            return NULL;
        }
        cpl_imagelist_subtract(imlist, imlist2);
        cpl_imagelist_delete(imlist2);
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode"); 
        return NULL;
    }

    /* Correct the detector linearity */
    if (detlin_a && detlin_b && detlin_c) {
        cpl_msg_info(cpl_func, "Correct for non-linearity");
        if (irplib_detlin_correct(imlist, detlin_a, detlin_b, detlin_c) == -1) {
            cpl_msg_error(cpl_func, "Cannot correct for non-linearity");
        }
    }

    /* Divide by the flatfield if one is provided */
    if (flat) {
        cpl_image * flat_im = cpl_image_load(flat, CPL_TYPE_UNSPECIFIED, 0, 0);
        cpl_error_code error = flat_im ? CPL_ERROR_NONE : cpl_error_get_code();

        cpl_msg_info(cpl_func, "Divide by the flat field");

        /* Threshold all pixels below the "mid-point" of FLT_MIN and 1, and
           all pixels above the "mid-point" of 1 and FLT_MAX. */
        error |= cpl_image_threshold(flat_im, sqrt(FLT_MIN), sqrt(FLT_MAX),
                                     1.0, 1.0);
        error |= cpl_imagelist_divide_image(imlist, flat_im);
        cpl_image_delete(flat_im);

        if (error) {
            cpl_imagelist_delete(imlist);
            (void)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                        "Flat field correction failed");
            return NULL;
        }
    }

    /* Different in chopping and non-chopping modes */
    if (isaac_img_zpoint_config.chopping == 1) {
        diffs = imlist;
        imlist = NULL;
    } else if (isaac_img_zpoint_config.chopping == 0) {
        /* Build the difference image list */
        niter = 2 * cpl_imagelist_get_size(imlist) - 2;
        diffs = cpl_imagelist_new();
        for (i=0; i<cpl_imagelist_get_size(imlist)-1; i++) {
            ima = cpl_image_subtract_create(
                    cpl_imagelist_get(imlist, i),
                    cpl_imagelist_get(imlist, i+1));
            cpl_imagelist_set(diffs, ima, 2*i);
            ima = cpl_image_subtract_create(
                    cpl_imagelist_get(imlist, i+1),
                    cpl_imagelist_get(imlist, i));
            cpl_imagelist_set(diffs, ima, 2*i+1);
        }
        cpl_imagelist_delete(imlist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "Cannot build the difference images");
            cpl_imagelist_delete(diffs);
            return NULL;
        }
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode - abort");
        cpl_imagelist_delete(imlist);
        return NULL;
    }
    return diffs;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Reduce one image
  @param    ilist   the input image list
  @param    pos     the objects positions
  @return   the newly allocated table with the following columns:
            POSX POSY ZPOINT FLUX BGD PEAK FWHMX FWHMY
            or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_img_zpoint_photom(
        cpl_imagelist       *   ilist,
        cpl_bivector        *   pos)
{
    cpl_table       *   tab;
    int                 nbima;
    double              r, r1, r2, mag, dit;
    cpl_image       *   ima;
    double          *   pos_x;
    double          *   pos_y;
    double              bgd, fl, zp, peak, fwhm_x, fwhm_y;
    int                 i;

    /* Test entries */
    if (ilist == NULL) return NULL;
    if (pos == NULL) return NULL;

    /* Initialise */
    nbima = cpl_imagelist_get_size(ilist);
    mag = isaac_img_zpoint_config.magnitude;
    dit = isaac_img_zpoint_config.dit;
    r = isaac_img_zpoint_config.phot_star_radius;
    r1 = isaac_img_zpoint_config.phot_bg_r1;
    r2 = isaac_img_zpoint_config.phot_bg_r2;

    /* Create the output table */
    tab = cpl_table_new(nbima);
    cpl_table_new_column(tab, "POSX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "POSY", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "ZPOINT", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FLUX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "PEAK", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "BGD", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FWHMX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(tab, "FWHMY", CPL_TYPE_DOUBLE);

    /* Loop on the images */
    pos_x = cpl_bivector_get_x_data(pos);
    pos_y = cpl_bivector_get_y_data(pos);
    for (i=0; i<nbima; i++) {
        /* Get the current image */
        ima = cpl_imagelist_get(ilist, i);
        /* Compute the photometry */
        /* Background */
        bgd = irplib_strehl_ring_background(ima, (int)(pos_x[i]), 
                (int)(pos_y[i]), (int)r1, (int)r2, IRPLIB_BG_METHOD_MEDIAN);
        /* Flux */
        fl = irplib_strehl_disk_flux(ima, 
                (int)(pos_x[i]), (int)(pos_y[i]), (int)r, bgd); 
        /* Zero Point */
        zp = mag + 2.5 * log10(fl) - 2.5 * log10(dit);
        cpl_msg_info(cpl_func, "Zero point nb %d:   %g", i+1, zp);
        /* Peak */
        peak = cpl_image_get_max_window(ima,
                (int)(pos_x[i]-5), (int)(pos_y[i]-5), 
                (int)(pos_x[i]+5), (int)(pos_y[i]+5)); 

        /* FWHM_X / FWHM_Y */
        if (cpl_image_get_fwhm(ima, (int)(pos_x[i]), (int)(pos_y[i]),
                &fwhm_x, &fwhm_y) != CPL_ERROR_NONE) {
            cpl_msg_debug(cpl_func, "Cannot compute FWHM for image %d", i+1);
            cpl_error_reset();
        }
                
        /* Put the results in the table */
        cpl_table_set_double(tab, "POSX", i, pos_x[i]);
        cpl_table_set_double(tab, "POSY", i, pos_y[i]);
        cpl_table_set_double(tab, "ZPOINT", i, zp);
        cpl_table_set_double(tab, "FLUX", i, fl);
        cpl_table_set_double(tab, "PEAK", i, peak);
        cpl_table_set_double(tab, "BGD", i, bgd);
        cpl_table_set_double(tab, "FWHMX", i, fwhm_x);
        cpl_table_set_double(tab, "FWHMY", i, fwhm_y);
    }
    return tab;
}

/*----------------------------------------------------------------------------*/
/*
  @brief    Compute the flux gradients
  @param    flux        a vector with 8 fluxes
  @param    positions   a bivector with the 8 positions
  @param    gradx       the x gradient
  @param    grady       the y gradient
  @param    graddx      the x gradient error
  @param    graddy      the y gradient error
  @return   0 if ok, -1 otherwise
     3    2
       1
     4    5
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_gradients(
        cpl_vector      *   flux,
        cpl_bivector    *   positions,
        double          *   gradx,
        double          *   grady,
        double          *   graddx,
        double          *   graddy) 
{
    int                 nflux;
    double              two, three, four, five, mean;
    double              two_pos_x, three_pos_x, four_pos_x, five_pos_x,
                        two_pos_y, three_pos_y, four_pos_y, five_pos_y;
    double          *   pflux;
    double          *   ppos_x;
    double          *   ppos_y;

    /* Test entries */
    if (flux == NULL) return -1;

    /* Initialise */
    nflux = cpl_vector_get_size(flux);
    if (nflux != 8) {
        *gradx = -1.0;
        *grady = -1.0;
        *graddx = -1.0;
        *graddy = -1.0;
        return 0;
    }
    mean = cpl_vector_get_mean(flux);

    /* Get the 4 positions fluxes */
    pflux = cpl_vector_get_data(flux);
    ppos_x = cpl_bivector_get_x_data(positions);
    ppos_y = cpl_bivector_get_y_data(positions);
    /* Flux */
    two = (pflux[1] + pflux[2]) / 2.0; 
    three = (pflux[3] + pflux[4]) / 2.0; 
    four = (pflux[5] + pflux[6]) / 2.0; 
    five = pflux[7];
    /* Positions */
    two_pos_x = (ppos_x[1] + ppos_x[2]) / 2.0;
    two_pos_y = (ppos_y[1] + ppos_y[2]) / 2.0;
    three_pos_x = (ppos_x[3] + ppos_x[4]) / 2.0; 
    three_pos_y = (ppos_y[3] + ppos_y[4]) / 2.0; 
    four_pos_x = (ppos_x[5] + ppos_x[6]) / 2.0; 
    four_pos_y = (ppos_y[5] + ppos_y[6]) / 2.0; 
    five_pos_x = ppos_x[7];
    five_pos_y = ppos_y[7];

    /* Compute grad */
    *gradx = (two + five - three - four) / 2;
    *grady = (three + two - four - five) / 2;

    /* Normalise to the nominal distance */
    *gradx *= (512 / ((two_pos_x + five_pos_x - three_pos_x - four_pos_x)/2));
    *grady *= (512 / ((three_pos_y + two_pos_y - four_pos_y - five_pos_y)/2));

    /* Compute error */
    *graddx = sqrt((two-five)*(two-five) + (three-four)*(three-four))/2.0;
    *graddy = sqrt((three-two)*(three-two) + (four-five)*(four-five))/2.0;
    
    /* Normalise */
    *gradx /= mean;
    *grady /= mean;
    *graddx /= mean;
    *graddy /= mean;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets from the headers
  @param    set     the input frame set
  @return   the newly allocated bivector 
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * isaac_img_zpoint_get_offsets(cpl_frameset * set) 
{
    int                     nframes;
    int                     noffsets;
    cpl_bivector        *   offs;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    double                  off_x, off_y;
    int                     i;

    /* Test entries */
    if (set == NULL) return NULL;
    
    /* Initialize */
    nframes = cpl_frameset_get_size(set);
    
    /* Different in chopping and non-chopping modes */
    if (isaac_img_zpoint_config.chopping == 1) {
        offs = isaac_get_offsets(set);
    } else if (isaac_img_zpoint_config.chopping == 0) {
        noffsets = 2 * (nframes-1);

        /* Check error code */
        if (cpl_error_get_code()) return NULL;
        
        /* Create the bivector */
        offs = cpl_bivector_new(noffsets);
        
        /* Fill it */
        for (i=0; i<nframes; i++) {
            cur_frame = cpl_frameset_get_frame(set, i);
            plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),0);
            off_x = isaac_pfits_get_cumoffsetx(plist);
            off_y = isaac_pfits_get_cumoffsety(plist);
            cpl_propertylist_delete(plist);
            if (i==0) {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i, off_y);
            } else if (i==nframes-1) {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i-1, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i-1, off_y);
            } else {
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i, off_y);
                cpl_vector_set(cpl_bivector_get_x(offs), 2*i-1, off_x);
                cpl_vector_set(cpl_bivector_get_y(offs), 2*i-1, off_y);
            }
        }
    } else {
        cpl_msg_error(cpl_func, "Unsupported mode - abort");
        return NULL;
    }

    /* Check if it went ok */
    if (cpl_error_get_code()) {
        cpl_bivector_delete(offs);
        cpl_msg_error(cpl_func, "Cannot find offsets in headers");
        return NULL;
    }
    return offs;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the standard star magnitude
  @param    stdstars    the standard stars catalog
  @param    frame       the input frame
  @param    mag         the computed magnitude
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_get_mag(
        const char  *       stdstars,
        double              ra,
        double              dec,
        isaac_band          band)
{
    double              star_mag;
    char                star_name[512];
    char                cat_name[512];
    char                star_type[32];
    int                 ret;

    /* Initialise */
    ret = -1;

    switch (band) {
        /* SW mode */
        case ISAAC_BAND_J:
        case ISAAC_BAND_H:
        case ISAAC_BAND_K:
        case ISAAC_BAND_KS:
            cpl_msg_info(cpl_func, "Try in LCO-Palomar");
            strcpy(cat_name, "LCO-Palomar");
            if ((ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                        isaac_std_band_name(band), "LCO-Palomar.txt", 
                        &star_mag, star_name, star_type, 2.0)) == -1) {
                cpl_msg_info(cpl_func, "Try in LCO-Palomar-NICMOS-Red-Stars");
                strcpy(cat_name, "LCO-Palomar-NICMOS-Red-Stars");
                if ((ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                            isaac_std_band_name(band),
                        "LCO-Palomar-NICMOS-Red-Stars.txt", &star_mag,
                        star_name, star_type, 2.0)) == -1) {
                    cpl_msg_info(cpl_func, "Try in all catalogs");
                    strcpy(cat_name, "all");
                    ret=irplib_stdstar_get_mag(stdstars, ra, dec,
                            isaac_std_band_name(band), "all", &star_mag,
                            star_name, star_type, 2.0);
                }
            }
            break;
        /* LW mode  */
        case ISAAC_BAND_L:
        case ISAAC_BAND_M:
            cpl_msg_info(cpl_func, "Try in ESO-VanDerBliek");
            strcpy(cat_name, "ESO-VanDerBliek");
            if ((ret=irplib_stdstar_get_mag(stdstars, ra, dec,
                        isaac_std_band_name(band), "ESO-VanDerBliek.txt", 
                        &star_mag, star_name, star_type, 2.0)) == -1) {
                cpl_msg_info(cpl_func, "Try in MSSSO-Photometric");
                strcpy(cat_name, "MSSSO-Photometric");
                if ((ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                            isaac_std_band_name(band), "MSSSO-Photometric.txt",
                            &star_mag, star_name, star_type, 2.0)) == -1) {
                    cpl_msg_info(cpl_func, "Try in MSSSO-Spectroscopic");
                    strcpy(cat_name, "MSSSO-Spectroscopic");
                    if ((ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                                isaac_std_band_name(band),
                            "MSSSO-Spectroscopic.txt", &star_mag,
                            star_name, star_type, 2.0)) == -1) {
                        cpl_msg_info(cpl_func, "Try in all catalogs");
                        strcpy(cat_name, "all");
                        ret=irplib_stdstar_get_mag(stdstars, ra, dec,
                                isaac_std_band_name(band), "all",
                                &star_mag, star_name, star_type, 2.0);
                    }
                }
            }
            break;
        default:
            cpl_msg_error(cpl_func, "cannot determine associated filter");
            return -1;
    }

    /* Special case: swap K and Ks if needed */
    if ((ret==-1) && (band==ISAAC_BAND_K)) {
        strcpy(cat_name, "all");
        ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                isaac_std_band_name(ISAAC_BAND_KS), "all", &star_mag,
                star_name, star_type, 2.0);
    } else if ((ret==-1) && (band==ISAAC_BAND_KS)) {
        strcpy(cat_name, "all");
        ret = irplib_stdstar_get_mag(stdstars, ra, dec,
                isaac_std_band_name(ISAAC_BAND_K), "all", &star_mag,
                star_name, star_type, 2.0);
    }

    /* If the magnitude still is not know, abort */
    if (ret == -1) {
        cpl_msg_error(cpl_func, "star magnitude not found in database");
        return -1;
    }

    /* Store results */
    isaac_img_zpoint_config.magnitude = star_mag;
    strcpy(isaac_img_zpoint_config.starname, star_name);
    strcpy(isaac_img_zpoint_config.sptype, star_type);
    strcpy(isaac_img_zpoint_config.catalog, cat_name);
    return 0;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    Save the recipe product on disk
  @param    tab         the table to save
  @param    check_im    the check image or NULL
  @param    raw         the raw frames
  @param    parlist     the parameter list
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */ 
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_save(
        cpl_table           *   tab,
        cpl_image           *   check_im,
        cpl_frameset        *   raw,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   paflist;
    cpl_propertylist    *   qclist;
    const cpl_frame           *   ref_frame;
    const char          *   sval;
    double                  hum, airm, extinction;

    /* Initialise */
    airm = 0;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return -1;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return -1;
    }
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_append_string(qclist, "ESO QC FILTER REF",
            isaac_std_band_name(isaac_img_zpoint_config.band));
    if (isaac_img_zpoint_compute_keywords(raw, &hum, &airm) != -1) {
        cpl_propertylist_append_double(qclist, "ESO QC AMBI RHUM AVG", hum);
        cpl_propertylist_append_double(qclist, "ESO QC AIRMASS", airm);
    }
    cpl_propertylist_delete(plist);
    /* Get extinction */
    switch (isaac_img_zpoint_config.band) {
        case ISAAC_BAND_J:    extinction = 0.09; break;
        case ISAAC_BAND_JS:   extinction = 0.05; break;
        case ISAAC_BAND_H:    extinction = 0.04; break;
        case ISAAC_BAND_KS:   extinction = 0.06; break;
        default:        extinction = 0.00; break;
    }
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINT",
            isaac_img_zpoint_config.zpoint);
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINT ATX0",
            isaac_img_zpoint_config.zpoint + airm * extinction);
    cpl_propertylist_append_double(qclist, "ESO QC ZPOINTRMS",
            isaac_img_zpoint_config.zpointrms);
    cpl_propertylist_append_double(qclist, "ESO QC FLUX MED",
            isaac_img_zpoint_config.flux_med);
    cpl_propertylist_append_string(qclist, "ESO QC STDNAME",
            isaac_img_zpoint_config.starname);
    cpl_propertylist_append_string(qclist, "ESO QC SPECTYPE",
            isaac_img_zpoint_config.sptype);
    cpl_propertylist_append_double(qclist, "ESO QC STARMAG",
            isaac_img_zpoint_config.magnitude);
    cpl_propertylist_append_string(qclist, "ESO QC CATNAME",
            isaac_img_zpoint_config.catalog);
    cpl_propertylist_append_double(qclist, "ESO QC GRADX",
            isaac_img_zpoint_config.gradx);
    cpl_propertylist_append_double(qclist, "ESO QC GRADY",
            isaac_img_zpoint_config.grady);
    cpl_propertylist_append_double(qclist, "ESO QC GRADDX",
            isaac_img_zpoint_config.graddx);
    cpl_propertylist_append_double(qclist, "ESO QC GRADDY",
            isaac_img_zpoint_config.graddy);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM MEAN",
            isaac_img_zpoint_config.fwhm_mean);

    /* Write the zpoint table  */
    irplib_dfs_save_table(set,
            parlist,
            set,
            tab,
            NULL,
            "isaac_img_zpoint",
            ISAAC_IMG_ZPOINT_TAB,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_zpoint.fits");
    /* Write the check image   */
    if (check_im) {
        irplib_dfs_save_image(set,
                parlist,
                set,
                check_im,
                CPL_BPP_IEEE_FLOAT,
                "isaac_img_zpoint",
                ISAAC_IMG_ZPOINT_CHECK,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "isaac_img_zpoint_check.fits");
    }
    
    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return -1;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|ESO TPL ID|DATE-OBS|MJD-OBS|ESO INS MODE|"
        "ESO OBS ID|ESO DET DIT|ESO INS PIXSCALE|RA|DEC)$", 0);
    cpl_propertylist_delete(plist);

    /* Get the QC params in qclist and keys for paf in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, "ESO QC", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_ZPOINT_TAB);

    /* Save the PAF file */
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_zpoint",
            paflist,
            "isaac_img_zpoint.paf");
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    Compute the median of the values of a key in several frames
  @param    set     the input frame set
  @return   0 if everything is ok, -1 otherwise
 */ 
/*----------------------------------------------------------------------------*/
static int isaac_img_zpoint_compute_keywords(
        cpl_frameset    *   set,
        double          *   hum,
        double          *   airm)
{
    int                     nframes;
    cpl_vector          *   hum_vec;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    int                     i;

    /* Test inputs  */
    if (set == NULL) return -1;

    /* Initialize */
    nframes = cpl_frameset_get_size(set);
    *airm   = 0.0;

    hum_vec = cpl_vector_new(nframes);

    for (i=0; i<nframes; i++) {
        if (cpl_error_get_code()) {
            cpl_vector_delete(hum_vec);
            return -1;
        }
        cur_frame = cpl_frameset_get_frame(set, i);
        plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
        if (i==0) *airm += isaac_pfits_get_airmass_start(plist);
        if (i==nframes-1) *airm += isaac_pfits_get_airmass_end(plist);
        cpl_vector_set(hum_vec,  i, isaac_pfits_get_humidity_level(plist));
        cpl_propertylist_delete(plist);
        if (cpl_error_get_code()) {
            cpl_vector_delete(hum_vec);
            cpl_error_reset();
            return -1;
        }
    }
    *hum  = cpl_vector_get_mean(hum_vec);
    *airm /= 2;

    /* Free and return */
    cpl_vector_delete(hum_vec);
    if (cpl_error_get_code()) return -1;
    return 0;
}


